#ifndef _PRBLAS_H
#define _PRBLAS_H

#include <algorithm>
#include <iostream>
#include <typeinfo>
#include <assert.h>
#include "rvector.h"
#include "rmatrix.h"
#include "thread_manager.h"

namespace rat {

// Index set for a thread.
typedef struct _iterset {
	int begin_;
	int end_;
	int stride_;
} iterset;

// Set up iteration sets for block distribution.
template<typename T>
void set_iterset_block(int nthreads, int n, T* task_param)
{
	int ichunk = (n + nthreads - 1) / nthreads;
	int s = 0;
	for (int i=0; i < nthreads; ++i)
	{
		task_param[i].is.begin_= s;
		task_param[i].is.end_ = (s + ichunk) > n ? n : s + ichunk;
		task_param[i].is.stride_ = 1;
		s += ichunk;
	}
}

// Set up iteration sets for cyclic distribution.
template<typename T>
void set_iterset_cyclic(int nthreads, int n, T* task_param)
{
//	std::cout << "set_iterset_cyclic(" << nthreads << "," << n << ")\n";
	for (int i=0; i < nthreads; ++i)
	{
		task_param[i].is.begin_= i;
		task_param[i].is.end_ = (i < n) ? i + ((n - i - 1)/nthreads+1)*nthreads : i;
		task_param[i].is.stride_ = nthreads;
	}
//	for (int i=0; i < nthreads; ++i) {
//		std::cout << i << ": [" << task_param[i].is.begin_ << ":" << task_param[i].is.end_ <<
//				":" << task_param[i].is.stride_ << "]\n";
//	}
}

//template<bool TRANS>
//struct trans
//{
//};
//
//const trans<false> N;	// No transpose
//const trans<true>  T;	// Transpose

/*
 *  xswap
 */

//template<typename T, template<typename> class T1, template<typename> class T2>
//void _Pswap(ivector<T1<T> >& x, ivector<T2<T> >& y)
//{
//	for (int i=0; i < x.len(); ++i)
//	{
//		std::swap(x[i],y[i]);
//	}
//}
//
//template<template<typename> class T1, template<typename> class T2>
//void pdswap(ivector<T1<double> >& x, ivector<T2<double> >& y)
//{
//	_Pswap<double,T1,T2>(x,y);
//}
//
//template<template<typename> class T1, template<typename> class T2>
//void prswap(ivector<T1<rational> >& x, ivector<T2<rational> >& y)
//{
//	for (int i=0; i < x.len(); ++i) {
//		rational::swap(x[i],y[i]);
//	}
//}

/*
 * xscal [parallel]
 */
template<typename T, template<typename> class Tx>
struct scal_param
{
	iterset is;
	T alpha_;
	ivector<Tx<T> >* x_;
};

template<typename T, template<typename> class Tx>
class scal_task
{
public:
	static scal_param<T,Tx> params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_scal(void* t)
	{
		scal_param<T,Tx>* param = static_cast<scal_param<T,Tx>*>(t);
#ifdef DBGPAR
		std::stringstream ss;
		ss << "partial_scal: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
#endif

		for (int i=param->is.begin_; i < param->is.end_; i+=param->is.stride_)
		{
			param->x_->operator[](i) *= param->alpha_;
		}
	}
};

template<typename T, template<typename> class Tx>
thread_manager::task_t scal_task<T,Tx>::tasks[thread_manager::num_threads];

template<typename T, template<typename> class Tx>
scal_param<T,Tx> scal_task<T,Tx>::params[thread_manager::num_threads];

template<typename T, template<typename> class VT>
void _Pscal(T const& alpha, ivector<VT<T> >& x)
{
	int nthreads = thread_manager::num_threads;

#ifdef BLOCKDIST
	set_iterset_block(nthreads,x.len(),scal_task<T,VT>::params);
#else
	set_iterset_cyclic(nthreads,x.len(),scal_task<T,VT>::params);
#endif

	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		scal_task<T,VT>::params[i].alpha_ = alpha;
		scal_task<T,VT>::params[i].x_ = &x;
		scal_task<T,VT>::tasks[i].f = scal_task<T,VT>::partial_scal;
		scal_task<T,VT>::tasks[i].params = &scal_task<T,VT>::params[i];
	}
	thread_manager::execute(scal_task<T,VT>::tasks,thread_manager::num_threads);
}

template<template<typename> class VT>
void pdscal(double const& alpha, ivector<VT<double> >& x)
{
	_Pscal<double,VT>(alpha,x);
}

template<template<typename> class VT>
void prscal(rational const& alpha, ivector<VT<rational> >& x)
{
	_Pscal<rational,VT>(alpha,x);
}

/*
 * xcopy [parallel]
 */
template<typename T, template<typename> class Tx, template<typename> class Ty>
struct copy_param
{
	iterset is;
	ivector<Tx<T> > const* x_;
	ivector<Ty<T> >* y_;
	T result_;
};

template<typename T, template<typename> class T1, template<typename> class T2>
class copy_task
{
public:
	static copy_param<T,T1,T2> copy_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_copy(void* t)
	{
		copy_param<T,T1,T2>* param = static_cast<copy_param<T,T1,T2>*>(t);
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->y_->operator [](i) = param->x_->operator [](i);
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2>
thread_manager::task_t copy_task<T,T1,T2>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2>
copy_param<T,T1,T2> copy_task<T,T1,T2>::copy_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2>
void _Pcopy(ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	assert(x.len()==y.len());
#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,x.len(),copy_task<T,T1,T2>::copy_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,x.len(),copy_task<T,T1,T2>::copy_params);
#endif
	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		copy_task<T,T1,T2>::copy_params[i].x_ = &x;
		copy_task<T,T1,T2>::copy_params[i].y_ = &y;
		copy_task<T,T1,T2>::tasks[i].f = copy_task<T,T1,T2>::partial_copy;
		copy_task<T,T1,T2>::tasks[i].params = &copy_task<T,T1,T2>::copy_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(copy_task<T,T1,T2>::tasks,thread_manager::num_threads);
}

template<template<typename> class T1, template<typename> class T2>
void pdcopy(ivector<T1<double> > const& x, ivector<T2<double> >& y)
{
	_Pcopy<double,T1,T2>(x,y);
}

template<template<typename> class T1, template<typename> class T2>
void prcopy(ivector<T1<rational> > const& x, ivector<T2<rational> >& y)
{
	_Pcopy<rational,T1,T2>(x,y);
}

/*
 * xvdiv [parallel]
 */
template<typename T, template<typename> class Tx, template<typename> class Ty>
struct vdiv_param
{
	iterset is;
	ivector<Tx<T> > const* x_;
	ivector<Ty<T> >* y_;
	T result_;
};

template<typename T, template<typename> class T1, template<typename> class T2>
class vdiv_task
{
public:
	static vdiv_param<T,T1,T2> vdiv_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_vdiv(void* t)
	{
		vdiv_param<T,T1,T2>* param = static_cast<vdiv_param<T,T1,T2>*>(t);
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->y_->operator [](i) = param->y_->operator [](i) / param->x_->operator [](i);
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2>
thread_manager::task_t vdiv_task<T,T1,T2>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2>
vdiv_param<T,T1,T2> vdiv_task<T,T1,T2>::vdiv_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2>
void _Pvdiv(ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	assert(x.len()==y.len());
#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,x.len(),vdiv_task<T,T1,T2>::vdiv_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,x.len(),vdiv_task<T,T1,T2>::vdiv_params);
#endif
	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		vdiv_task<T,T1,T2>::vdiv_params[i].x_ = &x;
		vdiv_task<T,T1,T2>::vdiv_params[i].y_ = &y;
		vdiv_task<T,T1,T2>::tasks[i].f = vdiv_task<T,T1,T2>::partial_vdiv;
		vdiv_task<T,T1,T2>::tasks[i].params = &vdiv_task<T,T1,T2>::vdiv_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(vdiv_task<T,T1,T2>::tasks,thread_manager::num_threads);
}

template<template<typename> class T1, template<typename> class T2>
void prvdiv(ivector<T1<rational> > const& x, ivector<T2<rational> >& y)
{
	_Pvdiv<rational,T1,T2>(x,y);
}

/*
 * xvdivs [parallel]   2015dec20  called by LDL of ratutil.cpp
 */
template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class Tz >
struct vdivs_param
{
	iterset is;
	ivector<Tx<T> > const* x_;
	ivector<Ty<T> > const* y_;
	ivector<Tz<T> >* z_;
	T result_;
};

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
class vdivs_task
{
public:
	static vdivs_param<T,T1,T2,T3> vdivs_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_vdivs(void* t)
	{
		vdivs_param<T,T1,T2,T3>* param = static_cast<vdivs_param<T,T1,T2,T3>*>(t);
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->z_->operator [](i) = param->y_->operator [](i) / param->x_->operator [](i);
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2, template<typename> class T3>
thread_manager::task_t vdivs_task<T,T1,T2,T3>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2, template<typename> class T3>
vdivs_param<T,T1,T2,T3> vdivs_task<T,T1,T2,T3>::vdivs_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
void _Pvdivs(ivector<T1<T> > const& x, ivector<T2<T> > const& y, ivector<T3<T> >& z)
{
//	assert(x.len()==y.len());
	assert(y.len()==z.len());   // matrix diagonal cannot be used foe vector length
#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,x.len(),vdivs_task<T,T1,T2,T3>::vdivs_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,x.len(),vdivs_task<T,T1,T2,T3>::vdivs_params);
#endif
	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		vdivs_task<T,T1,T2,T3>::vdivs_params[i].x_ = &x;
		vdivs_task<T,T1,T2,T3>::vdivs_params[i].y_ = &y;
		vdivs_task<T,T1,T2,T3>::vdivs_params[i].z_ = &z;
		vdivs_task<T,T1,T2,T3>::tasks[i].f = vdivs_task<T,T1,T2,T3>::partial_vdivs;
		vdivs_task<T,T1,T2,T3>::tasks[i].params = &vdivs_task<T,T1,T2,T3>::vdivs_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(vdivs_task<T,T1,T2,T3>::tasks,thread_manager::num_threads);
}

template<template<typename> class T1, template<typename> class T2, template<typename> class T3>
void prvdivs(ivector<T1<rational> > const& x, ivector<T2<rational> > const& y, ivector<T3<rational> >& z)
{
	_Pvdivs<rational,T1,T2,T3>(x,y,z);
}

/*
 * xvma [parallel]   2015dec27  called by hesfrb of sehfv3int.cpp
 */
template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class Tz >
struct vma_param
{
	iterset is;
	ivector<Tx<T> > const* x_;
	ivector<Ty<T> > const* y_;
	ivector<Tz<T> >* z_;
	T result_;
};

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
class vma_task
{
public:
	static vma_param<T,T1,T2,T3> vma_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_vma(void* t)
	{
		vma_param<T,T1,T2,T3>* param = static_cast<vma_param<T,T1,T2,T3>*>(t);
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->z_->operator [](i) += param->x_->operator [](i) * param->y_->operator [](i);
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2, template<typename> class T3>
thread_manager::task_t vma_task<T,T1,T2,T3>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2, template<typename> class T3>
vma_param<T,T1,T2,T3> vma_task<T,T1,T2,T3>::vma_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
void _Pvma(ivector<T1<T> > const& x, ivector<T2<T> > const& y, ivector<T3<T> >& z)
{
//	assert(x.len()==y.len());
	assert(y.len()==z.len());   // matrix diagonal cannot be used foe vector length
#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,x.len(),vma_task<T,T1,T2,T3>::vma_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,x.len(),vma_task<T,T1,T2,T3>::vma_params);
#endif
	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		vma_task<T,T1,T2,T3>::vma_params[i].x_ = &x;
		vma_task<T,T1,T2,T3>::vma_params[i].y_ = &y;
		vma_task<T,T1,T2,T3>::vma_params[i].z_ = &z;
		vma_task<T,T1,T2,T3>::tasks[i].f = vma_task<T,T1,T2,T3>::partial_vma;
		vma_task<T,T1,T2,T3>::tasks[i].params = &vma_task<T,T1,T2,T3>::vma_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(vma_task<T,T1,T2,T3>::tasks,thread_manager::num_threads);
}

template<template<typename> class T1, template<typename> class T2, template<typename> class T3>
void prvma(ivector<T1<rational> > const& x, ivector<T2<rational> > const& y, ivector<T3<rational> >& z)
{
	_Pvma<rational,T1,T2,T3>(x,y,z);
}


/*
 * xaxpy
 */
template<typename T, template<typename> class T1, template<typename> class T2>
struct axpy_param
{
	iterset is;
	T const* alpha_;
	ivector<T1<T> > const* x_;
	ivector<T2<T> >* y_;
};

template<typename T, template<typename> class T1, template<typename> class T2>
class axpy_task
{
public:
	static axpy_param<T,T1,T2> axpy_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_axpy(void* t)
	{
		axpy_param<T,T1,T2>* param = static_cast<axpy_param<T,T1,T2>*>(t);
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->y_->operator [](i) += *param->alpha_*param->x_->operator [](i);
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2>
thread_manager::task_t axpy_task<T,T1,T2>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2>
axpy_param<T,T1,T2> axpy_task<T,T1,T2>::axpy_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2>
void _Paxpy(T const& alpha, ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	assert(x.len()==y.len());
#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,x.len(),axpy_task<T,T1,T2>::axpy_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,x.len(),axpy_task<T,T1,T2>::axpy_params);
#endif
	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		axpy_task<T,T1,T2>::axpy_params[i].alpha_ = &alpha;
		axpy_task<T,T1,T2>::axpy_params[i].x_ = &x;
		axpy_task<T,T1,T2>::axpy_params[i].y_ = &y;
		axpy_task<T,T1,T2>::tasks[i].f = axpy_task<T,T1,T2>::partial_axpy;
		axpy_task<T,T1,T2>::tasks[i].params = &axpy_task<T,T1,T2>::axpy_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(axpy_task<T,T1,T2>::tasks,thread_manager::num_threads);
}

template<template<typename> class T1, template<typename> class T2>
void pdaxpy(double const& alpha, ivector<T1<double> > const& x, ivector<T2<double> >&y)
{
	_Paxpy<double,T1,T2>(alpha,x,y);
}

template<template<typename> class T1, template<typename> class T2>
void praxpy(rational const& alpha, ivector<T1<rational> > const& x, ivector<T2<rational> >&y)
{
	_Paxpy<rational,T1,T2>(alpha,x,y);
}

// dot [parallel]

template<typename T, template<typename> class T1, template<typename> class T2>
struct dot_param
{
	iterset is;
	ivector<T1<T> > const* x_;
	ivector<T2<T> > const* y_;
	T result_;
};

template<typename T, template<typename> class T1, template<typename> class T2>
class dot_task
{
public:
	static dot_param<T,T1,T2> dot_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_dot(void* t)
	{
		dot_param<T,T1,T2>* param = static_cast<dot_param<T,T1,T2>*>(t);
		param->result_ = T();
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_)
		{
			param->result_ += (param->x_->operator [](i)*param->y_->operator [](i));
		}
	}
};

template<typename T,template<typename> class T1, template<typename> class T2>
thread_manager::task_t dot_task<T,T1,T2>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class T1, template<typename> class T2>
dot_param<T,T1,T2> dot_task<T,T1,T2>::dot_params[thread_manager::num_threads];

template<typename T, template<typename> class T1, template<typename> class T2>
T _Pdot(const ivector<T1<T> >& a, const ivector<T2<T> >& b)
{
	assert(a.len()==b.len());
	int nthreads = thread_manager::num_threads;

#ifdef BLOCKDIST
	set_iterset_block(thread_manager::num_threads,a.len(),dot_task<T,T1,T2>::dot_params);
#else
	set_iterset_cyclic(thread_manager::num_threads,a.len(),dot_task<T,T1,T2>::dot_params);
#endif

	for (int i=0; i < thread_manager::num_threads; ++i)
	{
		dot_task<T,T1,T2>::dot_params[i].x_ = &a;
		dot_task<T,T1,T2>::dot_params[i].y_ = &b;
		dot_task<T,T1,T2>::tasks[i].f = dot_task<T,T1,T2>::partial_dot;
		dot_task<T,T1,T2>::tasks[i].params = &dot_task<T,T1,T2>::dot_params[i];
	}
	// Execute parallel tasks.
	thread_manager::execute(dot_task<T,T1,T2>::tasks,thread_manager::num_threads);
	// Reduce partial results.
	T prod = T();
	for (int i=0; i < nthreads; ++i)
	{
		prod = prod + dot_task<T,T1,T2>::dot_params[i].result_;
	}
	return prod;
}

template<template<typename> class T1, template<typename> class T2>
double pidot(const ivector<T1<int> >& a, const ivector<T2<int> >& b)
{
	return _Pdot<int,T1,T2>(a,b);
}

template<template<typename> class T1, template<typename> class T2>
double pddot(const ivector<T1<double> >& a, const ivector<T2<double> >& b)
{
	return _Pdot<double,T1,T2>(a,b);
}

template<template<typename> class T1, template<typename> class T2>
rational prdot(const ivector<T1<rational> >& a, const ivector<T2<rational> >& b)
{
	return _Pdot<rational,T1,T2>(a,b);
}

/*
 * asum [parallel]
 */

template<typename T, template<typename> class VT>
struct asum_param
{
	iterset is;
	ivector<VT<T> > const* x_;
	T result_;
};

template<typename T, template<typename> class VT>
class asum_task
{
public:
	static asum_param<T,VT> asum_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_asum(void* t)
	{
		asum_param<T,VT>* param = static_cast<asum_param<T,VT>*>(t);
		param->result_ = T();
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			T val = param->x_->operator[](i);
			param->result_ += (val > T() ? val : -val);
		}
	}
};

template<typename T,template<typename> class VT>
thread_manager::task_t asum_task<T,VT>::tasks[thread_manager::num_threads];

template<typename T,template<typename> class VT>
asum_param<T,VT> asum_task<T,VT>::asum_params[thread_manager::num_threads];

template<typename T, template<typename> class VT>
double _Pasum(ivector<VT<T> > const& x)
{
	int nthreads = thread_manager::num_threads;
#ifdef BLOCKDIST
	set_iterset_block(nthreads, x.len(), asum_task<T,VT>::asum_params);
#else
	set_iterset_cyclic(nthreads, x.len(), asum_task<T,VT>::asum_params);
#endif
	for (int i=0; i < nthreads; ++i) {
		asum_task<T,VT>::asum_params[i].x_ = &x;
		asum_task<T,VT>::tasks[i].f = &asum_task<T,VT>::partial_asum;
		asum_task<T,VT>::tasks[i].params = &asum_task<T,VT>::asum_params[i];
	}
	thread_manager::execute(asum_task<T,VT>::tasks,nthreads);
	T sum = T();
	for (int i=0; i< nthreads; ++i) {
		sum += asum_task<T,VT>::asum_params[i].result_;
	}
	return sum;
}

template<template<typename> class VT>
double pdasum(ivector<VT<double> > const& x)
{
	return _Pasum<double,VT>(x);
}

template<template<typename> class VT>
rational prasum(ivector<VT<rational> > const& x)
{
	return _Pasum<rational,VT>(x);
}

// ixdamax

//template<typename T, template<typename> class VT>
//int _Rixamax(ivector<VT<T> > const& x)
//{
//	int maxidx = 0;
//	T maxval = abs(x[maxidx]);
//	if (x.len() > 1) {
//		for (int i=0; i < x.len(); ++i) {
//			T val = abs(x[i]);
//			if (abs(val) > maxval) {
//				maxval = val;
//				maxidx = i;
//			}
//		}
//	}
//	return maxidx;
//}

//template<template<typename> class VT>
//int idamax(ivector<VT<double> > const& x)
//{
//	return _Rixamax<double,VT>(x);
//}
//
//template<template<typename> class VT>
//int iramax(ivector<VT<rational> > const& x)
//{
//	return _Rixamax<rational,VT>(x);
//}

/*
 *  dgemv [parallel]
 */
template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
struct gemv_param {
	iterset is;
	T const* alpha_;
	imatrix<TA<T> > const* A_;
	ivector<Tx<T> > const* x_;
	T const* beta_;
	ivector<Ty<T> >* y_;
};

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
class gemv_task
{
public:
	static gemv_param<T,TA,Tx,Ty> gemv_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_gemv0(void* t) {
		gemv_param<T,TA,Tx,Ty>* param = static_cast<gemv_param<T,TA,Tx,Ty>*>(t);
#ifdef DBGPAR
		std::stringstream ss;
		ss << "partial_gemv0: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
#endif
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			T dprod = T();
			for (int j=0; j < param->A_->columns(); ++j) {
				dprod = dprod + (*param->A_)[i][j]*(*param->x_)[j];
			}
			(*param->y_)[i] = *(param->alpha_) * dprod;
		}
	}
	static void partial_gemv(void* t) {
		gemv_param<T,TA,Tx,Ty>* param = static_cast<gemv_param<T,TA,Tx,Ty>*>(t);
		std::stringstream ss;
		ss << "partial_gemv: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			T dprod = T();
			for (int j=0; j < param->A_->columns(); ++j) {
				dprod = dprod + (*param->A_)[i][j] * (*param->x_)[j];
			}
			(*param->y_)[i] = (*param->alpha_) * dprod + (*param->beta_) * (*param->y_)[i];
		}
	}
};

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
thread_manager::task_t gemv_task<T,TA,Tx,Ty>::tasks[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
gemv_param<T,TA,Tx,Ty> gemv_task<T,TA,Tx,Ty>::gemv_params[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void _Pgemv(const T& alpha, const imatrix<TA<T> >& A, const ivector<Tx<T> >& x, const T& beta, ivector<Ty<T> >& y)
{
	assert(A.columns()==x.len() && A.rows()==y.len());
	int nthreads = thread_manager::num_threads;
#ifdef BLOCKDIST
	set_iterset_block(nthreads,A.rows(),gemv_task<T,TA,Tx,Ty>::gemv_params);
#else
	set_iterset_cyclic(nthreads,A.rows(),gemv_task<T,TA,Tx,Ty>::gemv_params);
#endif
	for (int i=0; i < nthreads; ++i) {
		gemv_task<T,TA,Tx,Ty>::gemv_params[i].alpha_ = &alpha;
		gemv_task<T,TA,Tx,Ty>::gemv_params[i].A_ = &A;
		gemv_task<T,TA,Tx,Ty>::gemv_params[i].x_ = &x;
		gemv_task<T,TA,Tx,Ty>::gemv_params[i].beta_ = &beta;
		gemv_task<T,TA,Tx,Ty>::gemv_params[i].y_ = &y;
		if (beta == T()) {
			gemv_task<T,TA,Tx,Ty>::tasks[i].f = gemv_task<T,TA,Tx,Ty>::partial_gemv0;
		}else{
			gemv_task<T,TA,Tx,Ty>::tasks[i].f = gemv_task<T,TA,Tx,Ty>::partial_gemv;
		}
		gemv_task<T,TA,Tx,Ty>::tasks[i].params = &gemv_task<T,TA,Tx,Ty>::gemv_params[i];
	}
	thread_manager::execute(gemv_task<T,TA,Tx,Ty>::tasks,nthreads);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void pdgemv(const double& alpha, const imatrix<TA<double> >& A, const ivector<Tx<double> >& x, const double& beta, ivector<Ty<double> >& y)
{
	_Pgemv<double,TA,Tx,Ty>(alpha,A,x,beta,y);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void prgemv(const rational& alpha, const imatrix<TA<rational> >& A, const ivector<Tx<rational> >& x, const rational& beta, ivector<Ty<rational> >& y)
{
	_Pgemv<rational,TA,Tx,Ty>(alpha,A,x,beta,y);
}


/*
 *  dgemtv [parallel]
 */
template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
struct gemtv_param {
	iterset is;
	T const* alpha_;
	imatrix<TA<T> > const* A_;
	ivector<Tx<T> > const* x_;
	T const* beta_;
	ivector<Ty<T> >* y_;
};

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
class gemtv_task
{
public:
	static gemtv_param<T,TA,Tx,Ty> gemtv_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_gemtv0(void* t) {
		gemtv_param<T,TA,Tx,Ty>* param = static_cast<gemtv_param<T,TA,Tx,Ty>*>(t);
#ifdef DBGPAR
		std::stringstream ss;
		ss << "partial_gemtv0: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
#endif
//		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
		for (int j=param->is.begin_; j < param->is.end_; j += param->is.stride_) {   // 2015dec16
			T dprod = T();
//			for (int j=0; j < param->A_->columns(); ++j) {
			for (int i=0; i < param->A_->rows(); ++i) {
//				dprod = dprod + (*param->A_)[i][j]*(*param->x_)[j];
				dprod = dprod + (*param->A_)[i][j]*(*param->x_)[i];  // 2015dec16
			}
//			(*param->y_)[i] = *(param->alpha_) * dprod;
			(*param->y_)[j] = *(param->alpha_) * dprod;  // 2015dec16
		}  // 2015dec16
	}
	static void partial_gemtv(void* t) {
		gemtv_param<T,TA,Tx,Ty>* param = static_cast<gemtv_param<T,TA,Tx,Ty>*>(t);
#ifdef DBGPAR
		std::stringstream ss;
		ss << "partial_gemtv: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
#endif
//		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
		for (int j=param->is.begin_; j < param->is.end_; j += param->is.stride_) {   // 2015dec16
			T dprod = T();
//			for (int j=0; j < param->A_->columns(); ++j) {
			for (int i=0; i < param->A_->rows(); ++i) {
//				dprod = dprod + (*param->A_)[i][j] * (*param->x_)[j];
				dprod = dprod + (*param->A_)[i][j]*(*param->x_)[i];  // 2015dec16
			}
//			(*param->y_)[i] = (*param->alpha_) * dprod + (*param->beta_) * (*param->y_)[i];
			(*param->y_)[j] = *(param->alpha_) * dprod + (*param->beta_) * (*param->y_)[j];  // 2015dec16
		}
	}
};

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
thread_manager::task_t gemtv_task<T,TA,Tx,Ty>::tasks[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
gemtv_param<T,TA,Tx,Ty> gemtv_task<T,TA,Tx,Ty>::gemtv_params[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void _Pgemtv(const T& alpha, const imatrix<TA<T> >& A, const ivector<Tx<T> >& x, const T& beta, ivector<Ty<T> >& y)
{
//	assert(A.columns()==x.len() && A.rows()==y.len());  // 2015dec16
	assert(A.rows()==x.len() && A.columns()==y.len());
	int nthreads = thread_manager::num_threads;
#ifdef BLOCKDIST
	set_iterset_block(nthreads,A.columns(),gemtv_task<T,TA,Tx,Ty>::gemtv_params);
#else
	set_iterset_cyclic(nthreads,A.columns(),gemtv_task<T,TA,Tx,Ty>::gemtv_params);
#endif
	for (int i=0; i < nthreads; ++i) {
		gemtv_task<T,TA,Tx,Ty>::gemtv_params[i].alpha_ = &alpha;
		gemtv_task<T,TA,Tx,Ty>::gemtv_params[i].A_ = &A;
		gemtv_task<T,TA,Tx,Ty>::gemtv_params[i].x_ = &x;
		gemtv_task<T,TA,Tx,Ty>::gemtv_params[i].beta_ = &beta;
		gemtv_task<T,TA,Tx,Ty>::gemtv_params[i].y_ = &y;
		if (beta == T()) {
			gemtv_task<T,TA,Tx,Ty>::tasks[i].f = gemtv_task<T,TA,Tx,Ty>::partial_gemtv0;
		}else{
			gemtv_task<T,TA,Tx,Ty>::tasks[i].f = gemtv_task<T,TA,Tx,Ty>::partial_gemtv;
		}
		gemtv_task<T,TA,Tx,Ty>::tasks[i].params = &gemtv_task<T,TA,Tx,Ty>::gemtv_params[i];
	}
	thread_manager::execute(gemtv_task<T,TA,Tx,Ty>::tasks,nthreads);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void pdgemtv(const double& alpha, const imatrix<TA<double> >& A, const ivector<Tx<double> >& x, const double& beta, ivector<Ty<double> >& y)
{
	_Pgemtv<double,TA,Tx,Ty>(alpha,A,x,beta,y);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void prgemtv(const rational& alpha, const imatrix<TA<rational> >& A, const ivector<Tx<rational> >& x, const rational& beta, ivector<Ty<rational> >& y)
{
	_Pgemtv<rational,TA,Tx,Ty>(alpha,A,x,beta,y);
}

/*
 * dger [parallel]
 */
template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
struct ger_param {
	iterset is;
	T const* alpha_;
	ivector<Tx<T> > const* x_;
	ivector<Ty<T> > const* y_;
	imatrix<TA<T> >* A_;
};

template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
class ger_task {
public:
	static ger_param<T,Tx,Ty,TA> ger_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_ger(void* t) {
		ger_param<T,Tx,Ty,TA>* param = static_cast<ger_param<T,Tx,Ty,TA>*>(t);
#ifdef DBGPAR
		std::stringstream ss;
		ss << "partial_ger: [" << param->is.begin_ << ":" << param->is.end_ << ":" << param->is.stride_ << "]\n";
		std::cout << ss.str();
#endif
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			for (int j=0; j < (*param->A_).columns(); ++j) {
				(*param->A_)[i][j] = (*param->alpha_) * (*param->x_)[i] * (*param->y_)[j] + (*param->A_)[i][j];
			}
		}
	}
};

template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
thread_manager::task_t ger_task<T,Tx,Ty,TA>::tasks[thread_manager::num_threads];

template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
ger_param<T,Tx,Ty,TA> ger_task<T,Tx,Ty,TA>::ger_params[thread_manager::num_threads];

template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void _Pger(const T& alpha, const ivector<Tx<T> >& x, const ivector<Ty<T> >& y, imatrix<TA<T> >& A)
{
	assert(x.len()==A.rows() && y.len()==A.columns());
	int nthreads = thread_manager::num_threads;
#ifdef BLOCKDIST
	set_iterset_block(nthreads,A.rows(),ger_task<T,Tx,Ty,TA>::ger_params);
#else
	set_iterset_cyclic(nthreads,A.rows(),ger_task<T,Tx,Ty,TA>::ger_params);
#endif
	for (int i=0; i < nthreads; ++i) {
		ger_task<T,Tx,Ty,TA>::ger_params[i].alpha_ = &alpha;
		ger_task<T,Tx,Ty,TA>::ger_params[i].x_ = &x;
		ger_task<T,Tx,Ty,TA>::ger_params[i].y_ = &y;
		ger_task<T,Tx,Ty,TA>::ger_params[i].A_ = &A;
		ger_task<T,Tx,Ty,TA>::tasks[i].f = ger_task<T,Tx,Ty,TA>::partial_ger;
		ger_task<T,Tx,Ty,TA>::tasks[i].params = &ger_task<T,Tx,Ty,TA>::ger_params[i];
	}
	thread_manager::execute(ger_task<T,Tx,Ty,TA>::tasks,nthreads);
}

template<template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void pdger(const double& alpha, const Tx<double>& x, const Ty<double>& y, TA<double>& A)
{
	_Pger<double,Tx,Ty,TA>(alpha,x,y,A);
}

template<template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void prger(const rational& alpha, const Tx<rational>& x, const Ty<rational>& y, TA<rational>& A)
{
	_Pger<rational,Tx,Ty,TA>(alpha,x,y,A);
}

/*
 *  gemm [parallel]
 */
template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
struct gemm_param {
	iterset is;
	T const* alpha_;
	imatrix<TA<T> > const* A_;
	imatrix<TB<T> > const* B_;
	T const* beta_;
	imatrix<TC<T> >* C_;
};
template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
class gemm_task {
public:
	static gemm_param<T,TA,TB,TC> gemm_params[thread_manager::num_threads];
	static thread_manager::task_t tasks[thread_manager::num_threads];
	static void partial_gemm1(void* t) {	// alpha==T() && beta==T()
		gemm_param<T,TA,TB,TC>* param = static_cast<ger_param<T,TA,TB,TC>*>(t);
		std::cout << "partial_gemm1\n";
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			for (int j=0; j < (*param->C_).columns(); ++j) {
				(*param->C_)[i][j] = T();
			}
		}
	}
	static void partial_gemm2(void* t) {	// alpha==T() && beta!=T()
		gemm_param<T,TA,TB,TC>* param = static_cast<ger_param<T,TA,TB,TC>*>(t);
		std::cout << "partial_gemm1\n";
		for (int i=param->is.begin_; i < param->is.end_; i += param->is.stride_) {
			for (int j=0; j < (*param->C_).columns(); ++j) {
				(*param->C_)[i][j] *= (*param->beta_);
			}
		}
	}
	static void partial_gemm3(void* t) {	// alpha != 0 && transa == 'N' && transb == 'N'
		int n, l;
		T temp;
		gemm_param<T,TA,TB,TC>* param = static_cast<gemm_param<T,TA,TB,TC>*>(t);
//		std::cout << "partial_gemm3\n";
		l = (*param->A_).columns();
		n = (*param->B_).columns();
		for (int i=param->is.begin_; i < param->is.end_; ++i) {
			for (int j=0; j < n; ++j) {
				temp = T();
				for (int k=0; k < l; ++k) {
					temp += (*param->A_)[i][k]*(*param->B_)[k][j];
				}
				(*param->C_)[i][j] = (*param->alpha_)*temp + (*param->beta_)*(*param->C_)[i][j];
			}
		}
	}
	static void partial_gemm4(void* t) {	// alpha != 0 && transa == 'N' && transb != 'N'
		int n, l;
		T temp;
		gemm_param<T,TA,TB,TC>* param = static_cast<gemm_param<T,TA,TB,TC>*>(t);
//		std::cout << "partial_gemm4\n";
		l = (*param->A_).columns();
		n = (*param->B_).rows();
		for (int i=param->is.begin_; i < param->is.end_; ++i) {
			for (int j=0; j < n; ++j) {
				temp = T();
				for (int k=0; k < l; ++k) {
					temp += (*param->A_)[i][k]*(*param->B_)[j][k];
				}
				(*param->C_)[i][j] = (*param->alpha_)*temp + (*param->beta_)*(*param->C_)[i][j];
			}
		}
	}
	static void partial_gemm5(void* t) {	// alpha != 0 && transa != 'N' && transb == 'N'
		int n, l;
		T temp;
		gemm_param<T,TA,TB,TC>* param = static_cast<gemm_param<T,TA,TB,TC>*>(t);
//		std::cout << "partial_gemm5\n";
		l = (*param->A_).rows();
		n = (*param->B_).columns();
		for (int i=param->is.begin_; i < param->is.end_; ++i) {
			for (int j=0; j < n; ++j) {
				temp = T();
				for (int k=0; k < l; ++k) {
					temp += (*param->A_)[k][i]*(*param->B_)[k][j];
				}
				(*param->C_)[i][j] = (*param->alpha_)*temp + (*param->beta_)*(*param->C_)[i][j];
			}
		}
	}
	static void partial_gemm6(void* t) {	// alpha != 0 && transa != 'N' && transb != 'N'
		int n, l;
		T temp;
		gemm_param<T,TA,TB,TC>* param = static_cast<gemm_param<T,TA,TB,TC>*>(t);
//		std::cout << "partial_gemm6\n";
		l = (*param->A_).rows();
		n = (*param->B_).rows();
		for (int i=param->is.begin_; i < param->is.end_; ++i) {
			for (int j=0; j < n; ++j) {
				temp = T();
				for (int k=0; k < l; ++k) {
					temp += (*param->A_)[k][i]*(*param->B_)[j][k];
				}
				(*param->C_)[i][j] = (*param->alpha_)*temp + (*param->beta_)*(*param->C_)[i][j];
			}
		}
	}
};

template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
thread_manager::task_t gemm_task<T,TA,TB,TC>::tasks[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
gemm_param<T,TA,TB,TC> gemm_task<T,TA,TB,TC>::gemm_params[thread_manager::num_threads];

template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
void _Pgemm(char transa, char transb, const T& alpha, const imatrix<TA<T> >& A, const imatrix<TB<T> >& B, const T& beta, imatrix<TC<T> >& C)
{
	int nthreads = thread_manager::num_threads;
	int m, n, l;
	void (*f)(void* t);
	if (alpha == T()) {
		//-----
#ifdef BLOCKDIST
		set_iterset_block(nthreads,C.rows(),gemm_task<T,TA,TB,TC>::gemm_params);
#else
		set_iterset_cyclic(nthreads,C.rows(),gemm_task<T,TA,TB,TC>::gemm_params);
#endif
		//-----
		if (beta == T()) {
			f = gemm_task<T,TA,TB,TC>::partial_gemm1;
		} else {
			f = gemm_task<T,TA,TB,TC>::partial_gemm2;
		}
	} else {
		T temp;
		if (transa == 'n' || transa == 'N'){
#ifdef BLOCKDIST
			set_iterset_block(nthreads,A.rows(),gemm_task<T,TA,TB,TC>::gemm_params);
#else
			set_iterset_cyclic(nthreads,A.rows(),gemm_task<T,TA,TB,TC>::gemm_params);
#endif
			if (transb == 'n' || transb == 'N'){
				f = gemm_task<T,TA,TB,TC>::partial_gemm3;
			} else {
				f = gemm_task<T,TA,TB,TC>::partial_gemm4;
			}
		} else {
#ifdef BLOCKDIST
			set_iterset_block(nthreads,A.columns(),gemm_task<T,TA,TB,TC>::gemm_params);
#else
			set_iterset_cyclic(nthreads,A.columns(),gemm_task<T,TA,TB,TC>::gemm_params);
#endif
			if (transb == 'n' || transb == 'N') {
				f = gemm_task<T,TA,TB,TC>::partial_gemm5;
			} else {
				f = gemm_task<T,TA,TB,TC>::partial_gemm6;
			}
		}
	}
	for (int i=0; i < nthreads; ++i) {
		gemm_task<T,TA,TB,TC>::gemm_params[i].alpha_ = &alpha;
		gemm_task<T,TA,TB,TC>::gemm_params[i].A_ = &A;
		gemm_task<T,TA,TB,TC>::gemm_params[i].B_ = &B;
		gemm_task<T,TA,TB,TC>::gemm_params[i].beta_ = &beta;
		gemm_task<T,TA,TB,TC>::gemm_params[i].C_ = &C;
		gemm_task<T,TA,TB,TC>::tasks[i].f = f;
		gemm_task<T,TA,TB,TC>::tasks[i].params = &gemm_task<T,TA,TB,TC>::gemm_params[i];
	}
	thread_manager::execute(gemm_task<T,TA,TB,TC>::tasks,nthreads);
}

template<template<typename> class TA, template<typename> class TB, template<typename> class TC>
void pdgemm(char transa, char transb, const double& alpha, const imatrix<TA<double> >& A, const imatrix<TB<double> >& B, const double& beta, imatrix<TC<double> >& C)
{
	_Pgemm<double,TA,TB,TC>(transa,transb,alpha,A,B,beta,C);
}

template<template<typename> class TA, template<typename> class TB, template<typename> class TC>
void prgemm(char transa, char transb, const rational& alpha, const imatrix<TA<rational> >& A, const imatrix<TB<rational> >& B, const rational& beta, imatrix<TC<rational> >& C)
{
	_Pgemm<rational,TA,TB,TC>(transa,transb,alpha,A,B,beta,C);
}

}	// namespace rational
#endif
