#ifndef _BLAS_H
#define _BLAS_H

#include <algorithm>
#include <iostream>
#include <typeinfo>
#include <assert.h>
#include "rvector.h"
#include "rmatrix.h"

namespace rat {

template<bool TRANS>
struct trans
{
};

const trans<false> N;	// No transpose
const trans<true>  T;	// Transpose

// xswap

template<typename T, template<typename> class T1, template<typename> class T2>
void _Rswap(ivector<T1<T> >& x, ivector<T2<T> >& y)
{
	for (int i=0; i < x.len(); ++i)
	{
		std::swap(x[i],y[i]);
	}
}

template<template<typename> class T1, template<typename> class T2>
void dswap(ivector<T1<double> >& x, ivector<T2<double> >& y)
{
	_Rswap<double,T1,T2>(x,y);
}

template<template<typename> class T1, template<typename> class T2>
void rswap(ivector<T1<rational> >& x, ivector<T2<rational> >& y)
{
	for (int i=0; i < x.len(); ++i) {
		rational::swap(x[i],y[i]);
	}
}

// xscal

template<typename T, template<typename> class VT>
void _Rscal(T const& alpha, ivector<VT<T> >& x)
{
	for (int i=0; i < x.len(); ++i) {
//		x[i] = alpha*x[i];
		x[i] *= alpha;
	}
}

template<template<typename> class VT>
void dscal(double const& alpha, ivector<VT<double> >& x)
{
	_Rscal<double,VT>(alpha,x);
}

template<template<typename> class VT>
void rscal(rational const& alpha, ivector<VT<rational> >& x)
{
	_Rscal<rational,VT>(alpha,x);
}

// dcopy()

template<typename T, template<typename> class T1, template<typename> class T2>
void _Rcopy(ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	for (size_t i=0; i < y.len(); ++i) {
		y[i] = x[i];
	}
}

template<template<typename> class T1, template<typename> class T2>
void dcopy(ivector<T1<double> > const& x, ivector<T2<double> >& y)
{
	_Rcopy<double,T1,T2>(x,y);
}

template<template<typename> class T1, template<typename> class T2>
void rcopy(ivector<T1<rational> > const& x, ivector<T2<rational> >& y)
{
	_Rcopy<rational,T1,T2>(x,y);
}

// xvdiv    Vector dividion, this entry is not in BLAS  2015nov07, 2015dec19

template<typename T, template<typename> class T1, template<typename> class T2>
void _Rvdiv(ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	for (int i=0; i < y.len(); ++i) {
		y[i] = y[i] / x[i];
	}
}

template<template<typename> class T1, template<typename> class T2>
void rvdiv(ivector<T1<rational> > const& x, ivector<T2<rational> >& y)
{
	_Rvdiv<rational,T1,T2>(x,y);
}

// xvdivs   Vector dividion, this entry is not in BLAS  2015dec20

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
void _Rvdivs(ivector<T1<T> > const& x, ivector<T2<T> > const& y, ivector<T3<T> >& z)
{
	for (int i=0; i < y.len(); ++i) {
		z[i] = y[i] / x[i];
	}
}

template<template<typename> class T1, template<typename> class T2, template<typename> class T3>
void rvdivs(ivector<T1<rational> > const& x, ivector<T2<rational> > const& y, ivector<T3<rational> >& z)
{
	_Rvdivs<rational,T1,T2,T3>(x,y,z);
}


// xvma   Vector Multiplication and Addition, this entry is not in BLAS  2015dec27

template<typename T, template<typename> class T1, template<typename> class T2, template<typename> class T3>
void _Rvma(ivector<T1<T> > const& x, ivector<T2<T> > const& y, ivector<T3<T> >& z)
{
	for (int i=0; i < y.len(); ++i) {
		z[i] += x[i] * y[i];
	}
}

template<template<typename> class T1, template<typename> class T2, template<typename> class T3>
void rvma(ivector<T1<rational> > const& x, ivector<T2<rational> > const& y, ivector<T3<rational> >& z)
{
	_Rvma<rational,T1,T2,T3>(x,y,z);
}



// xaxpy

template<typename T, template<typename> class T1, template<typename> class T2>
void _Raxpy(T const& alpha, ivector<T1<T> > const& x, ivector<T2<T> >& y)
{
	for (int i=0; i < y.len(); ++i) {
		y[i] += alpha*x[i];
	}
}

template<template<typename> class T1, template<typename> class T2>
void daxpy(double const& alpha, ivector<T1<double> > const& x, ivector<T2<double> >&y)
{
	_Raxpy<double,T1,T2>(alpha,x,y);
}

template<template<typename> class T1, template<typename> class T2>
void raxpy(rational const& alpha, ivector<T1<rational> > const& x, ivector<T2<rational> >&y)
{
	_Raxpy<rational,T1,T2>(alpha,x,y);
}

// xdot()

template<typename T, template<typename> class T1, template<typename> class T2>
T _Rdot(const ivector<T1<T> >& a, const ivector<T2<T> >& b)
{
	assert(a.len()==b.len());
	T dprod = T();
	for (size_t i=0; i < a.len(); ++i) {
		dprod += a[i]*b[i];
	}
	return dprod;
}

template<template<typename> class T1, template<typename> class T2>
double idot(const ivector<T1<int> >& a, const ivector<T2<int> >& b)
{
	return _Rdot<int,T1,T2>(a,b);
}

template<template<typename> class T1, template<typename> class T2>
double ddot(const ivector<T1<double> >& a, const ivector<T2<double> >& b)
{
	return _Rdot<double,T1,T2>(a,b);
}

template<template<typename> class T1, template<typename> class T2>
rational rdot(const ivector<T1<rational> >& a, const ivector<T2<rational> >& b)
{
	return _Rdot<rational,T1,T2>(a,b);
}

// xsum

template<typename T, template<typename> class VT>
T _Rasum(ivector<VT<T> > const& x)
{
	T res = T();
	for (int i=0; i < x.len(); ++i) {
		res += abs(x[i]);
	}
	return res;
}

template<template<typename> class VT>
double dasum(ivector<VT<double> > const& x)
{
	return _Rasum<double,VT>(x);
}

template<template<typename> class VT>
rational rasum(ivector<VT<rational> > const& x)
{
	return _Rasum<rational,VT>(x);
}

// ixdamax

template<typename T, template<typename> class VT>
int _Rixamax(ivector<VT<T> > const& x)
{
	int maxidx = 0;
	T maxval = abs(x[maxidx]);
	if (x.len() > 1) {
		for (int i=0; i < x.len(); ++i) {
			T val = abs(x[i]);
			if (abs(val) > maxval) {
				maxval = val;
				maxidx = i;
			}
		}
	}
	return maxidx;
}

template<template<typename> class VT>
int idamax(ivector<VT<double> > const& x)
{
	return _Rixamax<double,VT>(x);
}

template<template<typename> class VT>
int iramax(ivector<VT<rational> > const& x)
{
	return _Rixamax<rational,VT>(x);
}

// dgemv()

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void _Rgemv(const T& alpha, const imatrix<TA<T> >& A, const ivector<Tx<T> >& x, const T& beta, ivector<Ty<T> >& y)
{
	assert(A.columns()==x.len() && A.rows()==y.len());
	T dprod;
	if (beta == T()) {
		for (int i=0; i < A.rows(); ++i) {
			dprod = T();
			for (int j=0; j < A.columns(); ++j) {
				dprod = dprod + A[i][j]*x[j];
			}
			y[i] = alpha*dprod;
		}
	} else {
		for (int i=0; i < A.rows(); ++i) {
			dprod = T();
			for (int j=0; j < A.columns(); ++j) {
				dprod = dprod + A[i][j]*x[j];
			}
			y[i] = alpha*dprod + beta*y[i];
		}
	}
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void dgemv(const double& alpha, const imatrix<TA<double> >& A, const ivector<Tx<double> >& x, const double& beta, ivector<Ty<double> >& y)
{
	_Rgemv<double,TA,Tx,Ty>(alpha,A,x,beta,y);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void rgemv(const rational& alpha, const imatrix<TA<rational> >& A, const ivector<Tx<rational> >& x, const rational& beta, ivector<Ty<rational> >& y)
{
	_Rgemv<rational,TA,Tx,Ty>(alpha,A,x,beta,y);
}
// dgemtv()

template<typename T, template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void _Rgemtv(const T& alpha, const imatrix<TA<T> >& A, const ivector<Tx<T> >& x, const T& beta, ivector<Ty<T> >& y)
{
//	assert(A.columns()==x.len() && A.rows()==y.len());
	assert(A.rows()==x.len() && A.columns()==y.len());
	T dprod;
	if (beta == T()) {
//		for (int i=0; i < A.rows(); ++i) {
		for (int j=0; j < A.columns(); ++j) {
			dprod = T();
//			for (int j=0; j < A.columns(); ++j) {
			for (int i=0; i < A.rows(); ++i) {
//				dprod = dprod + A[i][j]*x[j];
				dprod = dprod + A[i][j]*x[i];
			}
//			y[i] = alpha*dprod;
			y[j] = alpha*dprod;
		}
	} else {
//		for (int i=0; i < A.rows(); ++i) {
		for (int j=0; j < A.columns(); ++j) {
			dprod = T();
//			for (int j=0; j < A.columns(); ++j) {
			for (int i=0; i < A.rows(); ++i) {
//				dprod = dprod + A[i][j]*x[j];
				dprod = dprod + A[i][j]*x[i];
			}
//			y[i] = alpha*dprod + beta*y[i];
			y[j] = alpha*dprod + beta*y[j];
		}
	}
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void dgemtv(const double& alpha, const imatrix<TA<double> >& A, const ivector<Tx<double> >& x, const double& beta, ivector<Ty<double> >& y)
{
	_Rgemtv<double,TA,Tx,Ty>(alpha,A,x,beta,y);
}

template<template<typename> class TA, template<typename> class Tx, template<typename> class Ty>
void rgemtv(const rational& alpha, const imatrix<TA<rational> >& A, const ivector<Tx<rational> >& x, const rational& beta, ivector<Ty<rational> >& y)
{
	_Rgemtv<rational,TA,Tx,Ty>(alpha,A,x,beta,y);
}


// dger

template<typename T, template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void _Rger(const T& alpha, const ivector<Tx<T> >& x, const ivector<Ty<T> >& y, imatrix<TA<T> >& A)
{
	assert(x.len()==A.rows() && y.len()==A.columns());
	for (int i=0; i < A.rows(); ++i) {
		for (int j=0; j < A.columns(); ++j) {
			A[i][j] = alpha*x[i]*y[j] + A[i][j];
		}
	}
}

template<template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void dger(const double& alpha, const Tx<double>& x, const Ty<double>& y, TA<double>& A)
{
	_Rger<double,Tx,Ty,TA>(alpha,x,y,A);
}

template<template<typename> class Tx, template<typename> class Ty, template<typename> class TA>
void rger(const rational& alpha, const Tx<rational>& x, const Ty<rational>& y, TA<rational>& A)
{
	_Rger<rational,Tx,Ty,TA>(alpha,x,y,A);
}

// dgemm

template<typename T, template<typename> class TA, template<typename> class TB, template<typename> class TC>
void _Rgemm(char transa, char transb, const T& alpha, const imatrix<TA<T> >& A, const imatrix<TB<T> >& B, const T& beta, imatrix<TC<T> >& C)
{
	int m, n, l;
	if (alpha == T()) {
		m = C.rows();
		n = C.columns();
		if (beta == T()) {
			for (int i=0; i < m; ++i) {
				for (int j=0; j < n; ++j) {
					C[i][j] = T();
				}
			}
		} else {
			for (int i=0; i < m; ++i) {
				for (int j=0; j < n; ++j) {
					C[i][j] *= beta;
				}
			}
		}
	} else {
		T temp;
		if (transa == 'n' || transa == 'N'){
			if (transb == 'n' || transb == 'N'){
				m = A.rows();
				l = A.columns();
				n = B.columns();
				for (int i=0; i < m; ++i) {
					for (int j=0; j < n; ++j) {
						temp = T();
						for (int k=0; k < l; ++k) {
							temp += A[i][k]*B[k][j];
						}
						C[i][j] = alpha*temp + beta*C[i][j];
					}
				}
			} else {
				m = A.rows();
				l = A.columns();
				n = B.rows();
				for (int i=0; i < m; ++i) {
					for (int j=0; j < n; ++j) {
						temp = T();
						for (int k=0; k < l; ++k) {
							temp += A[i][k]*B[j][k];
						}
						C[i][j] = alpha*temp + beta*C[i][j];
					}
				}
			}
		} else {
			if (transb == 'n' || transb == 'N') {
				m = A.columns();
				l = A.rows();
				n = B.columns();
				for (int i=0; i < m; ++i) {
					for (int j=0; j < n; ++j) {
						temp = T();
						for (int k=0; k < l; ++k) {
							temp += A[k][i]*B[k][j];
						}
						C[i][j] = alpha*temp + beta*C[i][j];
					}
				}
			} else {
				m = A.columns();
				l = A.rows();
				n = B.rows();
				for (int i=0; i < m; ++i) {
					for (int j=0; j < n; ++j) {
						temp = T();
						for (int k=0; k < l; ++k) {
							temp += A[k][i]*B[j][k];
						}
						C[i][j] = alpha*temp + beta*C[i][j];
					}
				}
			}
		}
	}
}

template<template<typename> class TA, template<typename> class TB, template<typename> class TC>
void dgemm(char transa, char transb, const double& alpha, const imatrix<TA<double> >& A, const imatrix<TB<double> >& B, const double& beta, imatrix<TC<double> >& C)
{
	_Rgemm<double,TA,TB,TC>(transa,transb,alpha,A,B,beta,C);
}

template<template<typename> class TA, template<typename> class TB, template<typename> class TC>
void rgemm(char transa, char transb, const rational& alpha, const imatrix<TA<rational> >& A, const imatrix<TB<rational> >& B, const rational& beta, imatrix<TC<rational> >& C)
{
	_Rgemm<rational,TA,TB,TC>(transa,transb,alpha,A,B,beta,C);
}

}	// namespace rational
#endif
