//============================================================================
// Name        : rough.cpp
// Author      : 
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C++, Ansi-style
//============================================================================

#include <iostream>
#include <sstream>
#include <cmath>
#include <cfloat>
#include <climits>
#include <limits>
#include <cassert>
#include "rough.h"

namespace rat
{
static const unsigned long long MANT_MASK = (1llu << (DBL_MANT_DIG - 1)) - 1;

const rough rough::ZERO((int32_t) 0,0);
const rough rough::ONE((int32_t) 1,0);
const rough rough::TWENTY((int32_t) 20,0);

union b64_t {
	double dbl;
	long long unsigned llu;
};

rough::intbits::intbits(rough const& r)
{
	b64_t t;
	t.dbl = r.significand();
	size_ = std::max(r.exponent()+1,0);
	intg_ = (t.llu & MANT_MASK);
	if (std::isnormal(t.dbl)) intg_ |= (1llu << (DBL_MANT_DIG-1));
}

bool rough::intbits::operator[](size_t pos) const
{
	if (DBL_MANT_DIG+pos < size_) {
		return 0;
	}
	return (intg_ >> (DBL_MANT_DIG+pos-size_)) & 1;
}

rough::fracbits::fracbits(rough const& r)
{
	union b64_t t;
	t.dbl = r.significand();
	size_ = std::max(DBL_MANT_DIG-1-r.exponent(),0);
	frac_ = t.llu & MANT_MASK;
	if (std::isnormal(t.dbl)) frac_ |= (1llu<<(DBL_MANT_DIG-1));
}

bool rough::fracbits::operator[](size_t pos) const
{
	if (DBL_MANT_DIG+pos < size_) {
		return 0;
	}
	return (frac_ >> (size_-pos-1)) & 1;
}

rough::decbuf::decbuf(size_t size, size_t dp)
{
	assert(size>=dp);
	size_ = size;
	dp_ = dp;
	buf_ = new char[size];
	assert(buf_ != 0);
}

rough::decbuf::~decbuf()
{
	delete[] buf_;
}

void rough::decbuf::clear()
{
	for (size_t i=0; i < size_; ++i) buf_[i] = 0;
}

rough::rough(double mant, int exp)
{
	int e;
	if (!std::isnormal(mant) && mant != 0.0) {
		mant = std::numeric_limits<double>::quiet_NaN();
		exp = 0;
	} else if (mant == 0.0) {
		exp = 0;
	} else if (std::abs(mant) < 1.0 || std::abs(mant) >= 2.0) {
		double m = frexp(mant, &e);
		if ((exp > 0 && e-1 > 0 && exp+e-1 <= 0) || (exp < 0 && e-1 < 0 && exp+e-1 > 0)) {
			mant = (m < 0 ? -1 : 1)*std::numeric_limits<double>::infinity();
			exp = 0;
		} else {
			mant = m * 2.0;
			exp += (e - 1);
		}
	}
	d_ = mant;
	e_ = exp;
}

rough::rough(const rough& r)
{
	d_ = r.d_;
	e_ = r.e_;
}

rough::intbits rough::getinteger() const
{
	return rough::intbits(*this);
}

rough::fracbits rough::getfractional() const
{
	return rough::fracbits(*this);
}

std::string rough::to_debug_string() const
{
	std::stringstream ss;
	ss.precision(17);
	ss << "{d: " << d_ << ", e: " << e_ <<"}";
	return ss.str();
}

double roughdec(const rough& f, int& m)
{
    int q = f.e_ / 10L; int r = f.e_ % 10L;
    double g = f.d_ * pow(2,r) * pow(1.024,q);
    m = 3*q;
    return g;
}

rough rough::abs() const{         // 2016sep29
	rough res(*this);
	rough rzero(0.0,0L);
	if( res < rzero ) res = -res;
	return res;
}

void rough::normalize()
{
	assert(d_ != std::numeric_limits<double>::infinity() &&
			d_ != -std::numeric_limits<double>::infinity() &&
			d_ != std::numeric_limits<double>::quiet_NaN());
	if (d_ != 0) {
		if (std::abs(d_) >= 2.0) {
			while (std::abs(d_) >= 2.0) {
				if (e_ == std::numeric_limits<long>::max()) {
					d_ = std::numeric_limits<double>::quiet_NaN();
					e_ = 0;
					break;
				}
				d_ /= 2.0;
				e_++;
			}
		} else if  (std::abs(d_) < 1.0) {
			while (std::abs(d_) < 1.0) {
				if (e_ == std::numeric_limits<long>::min()) {
					d_ = std::numeric_limits<double>::quiet_NaN();
					e_ = 0;
					break;
				}
				d_ *= 2.0;
				e_--;
			}
		}
	}
}

rough& rough::operator=(const rough& r) {
	d_ = r.d_;
	e_ = r.e_;
	return *this;
}

rough& rough::operator+=(const rough& r)
{
	if (r.d_ != 0.0) {
		rough t = r;
		if (e_ < t.e_) swap(*this,t);
		if (e_ == t.e_) {
			d_ += t.d_;
		} else if ((e_ - t.e_) < DBL_MANT_DIG) {
			d_ += (t.d_/std::pow(2.0,e_-t.e_));
		}
		if (d_ != 0) normalize();
	}
	return *this;
}

rough& rough::operator-=(const rough& r)
{
	if (r.d_ != 0.0) {
		if (d_ == 0.0) {
			d_ = -r.d_;
			e_ = r.e_;
		} else {
			rough t = -r;
			if (e_ < t.e_) swap(*this,t);
			if (e_ == t.e_) {
				d_ += t.d_;
			} else if ((e_ - t.e_) < DBL_MANT_DIG) {
				d_ += (t.d_/std::pow(2.0,e_-t.e_));
			}
			if (d_ != 0) normalize();
		}
	}
	return *this;
}

rough& rough::operator*=(const rough& r)
{
	if (std::isnormal(d_) && std::isnormal(r.d_)) {
		if ((e_ > 0 && r.e_ > 0 && e_ + r.e_ < 0) ||
		    (e_ < 0 && r.e_ < 0 && e_ + r.e_ > 0)) {
			d_ = std::numeric_limits<double>::quiet_NaN();
			e_ = 0;
		} else {
			d_ *= r.d_;
			e_ += r.e_;
			this->normalize();
		}
	} else if (!std::isnan(d_) && !std::isnan(r.d_)) {
		d_ = 0.0;
		e_ = 0;
	} else {
		d_ = std::numeric_limits<double>::quiet_NaN();
		e_ = 0;
	}
	return *this;
}

rough& rough::operator/=(const rough& r)
{
	if (d_ == 0 && std::isnormal(r.d_)) {
		d_ = 0.0;
		e_ = 0;
	} else if (std::isnormal(d_) && std::isnormal(r.d_)) {
		if ((e_ < 0 && r.e_ > 0 && e_-r.e_ >= 0) ||
			(e_ > 0 && r.e_ < 0 && e_-r.e_ <= 0)) {
			d_ = std::numeric_limits<double>::quiet_NaN();
			e_ = 0;
		} else {
			d_ /= r.d_;
			e_ -= r.e_;
			this->normalize();
		}
	} else {
		d_ = std::numeric_limits<double>::quiet_NaN();
		e_ = 0;
	}
	return *this;
}

// prefix ++
rough& rough::operator++()
{
	*this += rough(1.0,0);
	return *this;
}

// prefix --
rough& rough::operator--()
{
	*this -= rough(1.0,0);
	return *this;
}

// postfix ++
rough rough::operator++(int)
{
	rough r = *this;
	*this += rough(1.0,0);
	return r;
}

// postfix --
rough rough::operator--(int)
{
	rough r = *this;
	*this -= rough(1.0,0);
	return r;
}

// comparison operators
bool operator==(const rough& x, const rough& y)
{
	return (x.significand()==y.significand()) && (x.exponent()== y.exponent());
}

bool operator!=(const rough& x, const rough& y)
{
	return (x.significand()!=y.significand()) || (x.exponent()!= y.exponent());
}

bool operator>(const rough& x, const rough& y)
{
	if (x.d_ < 0.0 && y.d_ < 0.0) {
		if ((x.e_ < y.e_) || (x.e_ == y.e_ && x.d_ > y.d_))
			return true;
		else
			return false;
	} else if (x.d_ > 0.0 && y.d_ > 0.0) {
		if ((x.e_ > y.e_) || (x.e_ == y.e_ && x.d_ > y.d_))
			return true;
		else
			return false;
	} else if ((x.d_ == 0.0 && y.d_ < 0.0) || (x.d_ > 0.0 && y.d_ <= 0.0)) {
		return true;
	} else {
		return false;
	}
}

bool operator<(const rough& x, const rough& y)
{
	if (x.d_ < 0.0 && y.d_ < 0.0) {
		if ((x.e_ > y.e_) || (x.e_ == y.e_ && x.d_ < y.d_))
			return true;
		else
			return false;
	} else if (x.d_ > 0.0 && y.d_ > 0.0) {
		if ((x.e_ < y.e_) || (x.e_ == y.e_ && x.d_ < y.d_))
			return true;
		else
			return false;
	} else if ((x.d_ < 0.0 && y.d_ >= 0.0) || (x.d_ == 0.0 && y.d_ > 0.0)) {
		return true;
	} else {
		return false;
	}
}

bool operator>=(const rough& x, const rough& y)
{
	if (x.d_ < 0.0 && y.d_ < 0.0) {
		if ((x.e_ < y.e_) || (x.e_ == y.e_ && x.d_ >= y.d_))
			return true;
		else
			return false;
	} else if (x.d_ > 0.0 && y.d_ > 0.0) {
		if ((x.e_ > y.e_) || (x.e_ == y.e_ && x.d_ >= y.d_))
			return true;
		else
			return false;
	} else if (x.d_ >= 0.0 && y.d_ <= 0.0) {
		return true;
	} else {
		return false;
	}
}

bool operator<=(const rough& x, const rough& y)
{
	if (x.d_ < 0.0 && y.d_ < 0.0) {
		if ((x.e_ > y.e_) || (x.e_ == y.e_ && x.d_ <= y.d_))
			return true;
		else
			return false;
	} else if (x.d_ > 0.0 && y.d_ > 0.0) {
		if ((x.e_ < y.e_) || (x.e_ == y.e_ && x.d_ <= y.d_))
			return true;
		else
			return false;
	} else if (x.d_ <= 0.0 && y.d_ >= 0.0) {
		return true;
	} else {
		return false;
	}}

void rough::decbuf::insert_float(std::ostringstream& ss, size_t m, size_t l) {
//	ss << "insert_float: {m: " << m << ", l:" << l << "} ";
	ss << static_cast<int>(buf_[m-1]);
	if ((m-l) > 1) {
		ss << ".";
		for (size_t i=m-1; i > l; --i) ss << static_cast<int>(buf_[i-1]);
	}
	ss << 'e' << (m > dp_ ? '+' : '-') << std::setfill('0') << std::setw(3) <<
			(m > dp_ ? m-1-dp_ : dp_-m+1);
}

void rough::decbuf::insert_fixed(std::ostringstream& ss, size_t m, size_t l) {
//	ss << "insert_fixed: {m: " << m << ", l:" << l << "} ";
	if (m > dp_) {
		for(size_t i=m; i > dp_; --i) ss << static_cast<int>(buf_[i-1]);
		if (l < dp_) {
			ss << '.';
			for (size_t i=dp_; i > l; --i) ss << static_cast<int>(buf_[i-1]);
		}
	} else {
		ss << "0.";
		if (m < dp_) {
			for(size_t i=dp_; i > m; --i) ss << '0';
		}
		for (size_t i=m; i > l; --i) ss << static_cast<int>(buf_[i-1]);
	}
}

void rough::decbuf::insert_debug(size_t m, size_t l) {
	size_t k = std::max(dp_,l);
	std::cout << "{sz: " << size_ << ", dp: " << dp_ << ", m: " << m << ", l: " << l << "} ";
	for (size_t i=size_; i > k; --i) {
		if (i > m) {
			std::cout << static_cast<char>(buf_[i-1]+'a');
		} else {
			std::cout << static_cast<int>(buf_[i-1]);
		}
	}
	if (l < dp_) {
		std::cout << '.';
		for (size_t i=dp_; i > 0; --i) {
			if ( i<= m && i > l) {
				std::cout << static_cast<int>(buf_[i-1]);
			} else {
				std::cout << static_cast<char>(buf_[i-1]+'a');
			}
		}
	}
}
}	// end rat
