/*
 * rough.h
 *
 *  Created on: 2016/07/25
 *      Author: Osamu
 */

#ifndef ROUGH_H_
#define ROUGH_H_
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <cassert>
#include <cmath>
#include <limits>

namespace rat
{

class rough
{
	double d_;
	int e_;

	// some private non-member functions
	static double pos_inf() {
		return std::numeric_limits<double>::infinity();
	}
	static double neg_inf() {
		return -std::numeric_limits<double>::infinity();
	}
	static double qnan() {
		return std::numeric_limits<double>::quiet_NaN();
	}
	static bool is_pos_inf(double x) {
		return x == std::numeric_limits<double>::infinity();
	}
	static bool is_neg_inf(double x) {
		return x == -std::numeric_limits<double>::infinity();
	}
public:
	// Constants
	static const rough ZERO;
	static const rough ONE;
	static const rough TWENTY;

	class intbits
	{
		size_t size_;
		unsigned long long intg_;
	public:
		intbits(rough const&);
		size_t size() const { return size_; }
		bool operator[](size_t pos) const;
	};
	class fracbits
	{
		size_t size_;
		unsigned long long frac_;
	public:
		fracbits(rough const&);
		size_t size() const { return size_; }
		bool operator[](size_t pos) const;
	};
	class decbuf
	{
		size_t size_;	// the number of decimal digits
		size_t dp_;		// the position of decimal point
		char* buf_;
	public:
		decbuf(size_t size, size_t dp);
		~decbuf();
		size_t size() const { return size_; }
		size_t dp() const { return dp_; }
		void clear();
		char const& operator[](size_t pos) const {
			assert(pos<size_);
			return buf_[pos];
		}
		char& operator[](size_t pos) {
			assert(pos<size_);
			return buf_[pos];
		}
		char imul2add(char d) {
			char carry = 0;
			for (size_t j=dp_; j <size_; ++j) {
				char t = (buf_[j] << 1) + carry;
				if (t >= 10) {
					carry = 1;
					t -= 10;
				} else {
					carry = 0;
				}
				buf_[j] = t;
			}
			buf_[dp_] += d;
			return carry;
		}
		char fdiv2add(char d) {
			char dcarry = 0;
			for (size_t j=dp_; j > 0; --j) {
				char t = buf_[j-1] + dcarry;
				if ((t & 1) != 0) {
					dcarry = 10;
				} else {
					dcarry = 0;
				}
				buf_[j-1] = (t >> 1);
			}
			buf_[dp_-1] += d;
			return dcarry;
		}
		char round(size_t m, size_t l) {
			char carry=0;
			if (l != 0 && buf_[l-1] >= 5) {
				carry = 1;
				for(size_t i=l; carry !=0 && i < m; ++i) {
					char t = buf_[i] + carry;
					buf_[i] = t % 10;
					carry = t/10;
				}
			}
			return carry;
		}
		size_t suppress(size_t m, size_t l) {
			while(l < m && buf_[l]==0) ++l;
			return l;
		}
		void insert_float(std::ostringstream& ss, size_t m, size_t l);
		void insert_fixed(std::ostringstream& ss, size_t m, size_t l);

		// for debugging
		void insert_debug(size_t m, size_t l);
	};
	rough(): d_(0.0), e_(0) {}
	rough(double,int);
	rough(const rough&);
	virtual ~rough() {}
	double significand() const { return d_; }
	int exponent() const { return e_; }
	rough::intbits getinteger() const;
	rough::fracbits getfractional() const;
	void normalize();
	static bool isnan(const rough x) { return std::isnan(x.d_); }
	static bool ininf(const rough x) { return is_pos_inf(x.d_) || is_neg_inf(x.d_); }

	// assignment operators
	rough& operator=(const rough&);
	rough& operator+=(const rough&);
	rough& operator-=(const rough&);
	rough& operator*=(const rough&);
	rough& operator/=(const rough&);

	friend bool operator==(const rough& x, const rough& y);
	friend bool operator!=(const rough& x, const rough& y);
	friend bool operator>(const rough& x, const rough& y);
	friend bool operator<(const rough& x, const rough& y);
	friend bool operator>=(const rough& x, const rough& y);
	friend bool operator<=(const rough& x, const rough& y);

	// unary minus operator
	inline rough operator-() const {
		return rough(-d_,e_);
	}

	// increment/decrement operators
	rough& operator++();	// prefix ++
	rough& operator--();	// prefix --
	rough operator++(int);	// postfix ++
	rough operator--(int);	// postfix --

	// swap
	inline static void swap(rough& a, rough& b)
	{
		std::swap(a.d_,b.d_);
		std::swap(a.e_,b.e_);
	}

	// misc.
	friend double roughdec(const rough&, int& m);
	rough abs() const;             // 2016sep29


	// for debugging
	std::string to_debug_string() const;
};

// arithmetic operators
inline rough operator+(const rough& x, const rough& y)
{
	rough t = x;
	t += y;
	return t;
}
inline rough operator-(const rough& x, const rough& y)
{
	rough t = x;
	t -= y;
	return t;
}
inline rough operator*(const rough& x, const rough& y)
{
	rough t = x;
	t *= y;
	return t;
}
inline rough operator/(const rough& x, const rough& y)
{
	rough t = x;
	t /= y;
	return t;
}

template <typename charT, typename traits>
inline
std::basic_ostream<charT,traits>&
operator << (std::basic_ostream<charT,traits>& os, const rough& r)
{
	size_t prec = os.precision();
	rough::intbits n = r.getinteger();
	rough::fracbits f = r.getfractional();
	double const Log2 = 0.30103;
	size_t const idigits = static_cast<size_t>(ceil(Log2*n.size())+1);
	size_t const fdigits = std::max(static_cast<size_t>(ceil(Log2*f.size()+10.0)),prec+10);
	rough::decbuf dd(idigits+fdigits,fdigits);
	dd.clear();

	// Convert to fixed decimal
	if (n.size() > 0) {
		dd[fdigits] = n[n.size()-1];
		for (size_t i=n.size()-1; i > 0; --i) {
			dd.imul2add(n[i-1]);
		}
	}
	if (f.size() > 0) {
		dd[fdigits-1] = f[f.size()-1]*5;
		for (size_t i=f.size()-1; i > 0; --i) {
			dd.fdiv2add(f[i-1]*5);
		}
	}
	std::basic_ostringstream<charT,traits> ss;
	ss.copyfmt(os);
	ss.width(0);

	size_t msd, lsd;
	char carry;

	// find most significant digit position
	for (msd=dd.size(); msd > 0; --msd) {
		if (dd[msd-1] != 0) break;
	}
	lsd = msd < prec ? 0 : msd - prec;
	if (r.significand() < 0.0) {
		ss << '-';
	}
	if (msd == 0) {
		ss << '0';
	} else {
		// round up.
		carry = dd.round(msd,lsd);
		if (carry > 0) {
			lsd = msd;
			++msd;
			dd[msd-1] = carry;
		}

		// suppress trailing zeros.
		if (lsd < dd.dp()) {
			lsd = dd.suppress(dd.dp(),lsd);
		}
		// insert formatted string
		if (msd > dd.dp()+prec) {
			lsd = dd.suppress(msd,lsd);
			dd.insert_float(ss,msd,lsd);
		} else if (msd == dd.dp()+prec && lsd >= dd.dp()) {
			dd.insert_fixed(ss,msd,lsd);
		} else if (msd > dd.dp() && lsd < dd.dp()) {
			dd.insert_fixed(ss,msd,lsd);
		} else if (msd == dd.dp()) {
			dd.insert_fixed(ss,msd,lsd);
		} else {
			if (dd.dp()-lsd+2 < msd-lsd+1+5) {
				dd.insert_fixed(ss,msd,lsd);
			} else {
				dd.insert_float(ss,msd,lsd);
			}
		}
	}
	os << ss.str();
	return os;
}
}
#endif
